import re

def slugify(value):
    value = re.sub(r'[^\w]+', '_', value.lower()).strip('_')
    if value[:1].isdigit():
        value = '_' + value
    return value

# A single variable (enclosed by the template)
class Var(object):
    STRUCT_NAME = 'vars'
    CSIZES = {
        # This array doesn't have to be exact, it's only used for sorting
        # struct members to save a few bytes of memory here and there
        'int':      4,
        'unsigned': 4,
        'float':    4,
        'ident_t':  2,
        'uint8_t':  1,
        'bool':     1,
    }

    def __init__(self, ctype, expr, name, csize=0, linenr=0):
        self.ctype  = ctype
        self.csize  = csize or Var.CSIZES[ctype]
        self.expr   = expr
        self.name   = name
        self.linenr = linenr

    def __str__(self):
        return f'{Var.STRUCT_NAME}.{self.name}'

def is_literal(expr):
    return expr.isnumeric() or expr in ['true', 'false']

# A (deduplicated) set of variables
class VarSet(object):
    def __init__(self):
        self.varmap = {} # expr -> cvar

    def __iter__(self):
        # Sort from largest to smallest variable to optimize struct padding
        yield from sorted(self.varmap.values(),
            reverse=True,
            key=lambda v: v.csize,
        )

    def __bool__(self):
        return True if self.varmap else False

    def add_var_raw(self, var):
        # Re-use existing entry for identical expression/type pairs
        if old := self.varmap.get(var.expr):
            if var.ctype != old.ctype:
                raise SyntaxError(f'Conflicting types for expression {var.expr}, '
                                  f'got {var.ctype}, expected {old.ctype}')
            assert old.name == var.name
            return old

        names = [ v.name for v in self.varmap.values() ]
        while var.name in names:
            var.name += '_'
        self.varmap[var.expr] = var
        return var

    # Returns the added variable
    def add_var(self, ctype, expr, name=None, linenr=0):
        assert expr
        expr = expr.strip()
        if is_literal(expr):
            return expr
        name = name or slugify(expr)

        var = Var(ctype, expr=expr, name=name, linenr=linenr)
        return self.add_var_raw(var)

    def merge(self, other):
        for var in other:
            self.add_var_raw(var)



#!/usr/bin/env python3

import sys
import subprocess

infilename, outfilename, source_dir, project_version_pretty = sys.argv[1:]

try:
    proc = subprocess.run(['git', 'describe', '--dirty'], cwd=source_dir,
                          capture_output=True, text=True)
    proc.check_returncode()
except (FileNotFoundError, subprocess.CalledProcessError):
    # No git or no repo. Hopefully a release tarball.
    version = project_version_pretty
else:
    version = '{} ({})'.format(project_version_pretty, proc.stdout.strip())

with open(infilename, 'r') as infile:
    output = infile.read().replace('@buildver@', version)
# Avoid touching file (triggering recompilation) if it's already up to date.
try:
    with open(outfilename, 'r') as outfile:
        write_output = outfile.read() != output
except FileNotFoundError:
    write_output = True
if write_output:
    with open(outfilename, 'w') as outfile:
        outfile.write(output)



import jinja2
import os.path

TEMPLATEDIR = os.path.dirname(__file__) + '/templates'
TEMPLATES = jinja2.Environment(
    loader        = jinja2.FileSystemLoader(searchpath=TEMPLATEDIR),
    lstrip_blocks = True,
    trim_blocks   = True,
)

GLSL_BLOCK_TEMPLATE = TEMPLATES.get_template('glsl_block.c.j2')
FUNCTION_TEMPLATE   = TEMPLATES.get_template('function.c.j2')
CALL_TEMPLATE       = TEMPLATES.get_template('call.c.j2')
STRUCT_TEMPLATE     = TEMPLATES.get_template('struct.c.j2')



#!/usr/bin/env python3

import sys
import argparse

from macros import Macro

parser = argparse.ArgumentParser()
parser.add_argument('input')
parser.add_argument('output')
parser.add_argument('-s', '--strip', default=False, action='store_true')
args = parser.parse_args()

with open(args.input, encoding='utf-8') as infile:
    with open(args.output, 'w', encoding='utf-8') as outfile:
        for line in Macro.process_file(infile, strip=args.strip):
            outfile.write(line)



import re

from templates import GLSL_BLOCK_TEMPLATE
from variables import VarSet, slugify

VAR_PATTERN = re.compile(flags=re.VERBOSE, pattern=r'''
    # long form ${ ... } syntax
    \${ (?:\s*(?P<type>(?:              # optional type prefix
            ident                       # identifiers (always dynamic)
          | (?:(?:const|dynamic)\s+)?   # optional const/dynamic modifiers
                (?:float|u?int)         # base type
          | swizzle                     # swizzle mask
          | (?:i|u)?vecType             # vector type (for mask)
        )):)?
        (?P<expr>[^{}]+)
      }
|   \$(?P<name>\w+) # reference to captured variable
|   @(?P<var>\w+)   # reference to locally defined var
''')

class FmtSpec(object):
    def __init__(self, ctype='ident_t', fmtstr='_%hx',
                 wrap_expr=lambda name, expr: expr,
                 fmt_expr=lambda name: name):
        self.ctype     = ctype
        self.fmtstr    = fmtstr
        self.wrap_expr = wrap_expr
        self.fmt_expr  = fmt_expr

    @staticmethod
    def wrap_var(type, dynamic=False):
        if dynamic:
            return lambda name, expr: f'sh_var_{type}(sh, "{name}", {expr}, true)'
        else:
            return lambda name, expr: f'sh_const_{type}(sh, "{name}", {expr})'

    @staticmethod
    def wrap_fn(fn):
        return lambda name: f'{fn}({name})'

VAR_TYPES = {
    # identifiers: get mapped as-is
    'ident':            FmtSpec(),

    # normal variables: get mapped as shader constants
    'int':              FmtSpec(wrap_expr=FmtSpec.wrap_var('int')),
    'uint':             FmtSpec(wrap_expr=FmtSpec.wrap_var('uint')),
    'float':            FmtSpec(wrap_expr=FmtSpec.wrap_var('float')),

    # constant variables: get printed directly into the source code
    'const int':        FmtSpec(ctype='int',          fmtstr='%d'),
    'const uint':       FmtSpec(ctype='unsigned',     fmtstr='uint(%u)'),
    'const float':      FmtSpec(ctype='float',        fmtstr='float(%f)'),

    # dynamic variables: get loaded as shader variables
    'dynamic int':      FmtSpec(wrap_expr=FmtSpec.wrap_var('int', dynamic=True)),
    'dynamic uint':     FmtSpec(wrap_expr=FmtSpec.wrap_var('uint', dynamic=True)),
    'dynamic float':    FmtSpec(wrap_expr=FmtSpec.wrap_var('float', dynamic=True)),

    # component mask types
    'swizzle':          FmtSpec(ctype='uint8_t', fmtstr='%s', fmt_expr=FmtSpec.wrap_fn('sh_swizzle')),
    'ivecType':         FmtSpec(ctype='uint8_t', fmtstr='%s', fmt_expr=FmtSpec.wrap_fn('sh_float_type')),
    'uvecType':         FmtSpec(ctype='uint8_t', fmtstr='%s', fmt_expr=FmtSpec.wrap_fn('sh_float_type')),
    'vecType':          FmtSpec(ctype='uint8_t', fmtstr='%s', fmt_expr=FmtSpec.wrap_fn('sh_float_type')),
}

def stringify(value, strip):
    if strip:
        value = re.sub(r'(?:\/\*[^\*]*\*\/|\/\/[^\n]+|^\s*)', '', value)
    return '"' + value.replace('\\', '\\\\').replace('"', '\\"') + '\\n"'

def commentify(value, strip):
    if strip:
        return ''
    return '/*' + value.replace('/*', '[[').replace('*/', ']]') + '*/'

# Represents a statement + its enclosed variables
class Statement(object):
    def __init__(self, linenr=0):
        super().__init__()
        self.linenr = linenr
        self.vars = VarSet()

    def add_var(self, ctype, expr, name=None):
        return self.vars.add_var(ctype, expr, name, self.linenr)

    def render(self):
        raise NotImplementedError

    @staticmethod
    def parse(text_orig, **kwargs):
        raise NotImplementedError

# Represents a single line of GLSL
class GLSLLine(Statement):
    class GLSLVar(object): # variable reference
        def __init__(self, fmt, var):
            self.fmt = fmt
            self.var = var

    def __init__(self, text, strip=False, **kwargs):
        super().__init__(**kwargs)
        self.refs = []
        self.strip = strip

        # produce two versions of line, one for printf() and one for append()
        text = text.rstrip()
        self.rawstr = stringify(text, strip)
        self.fmtstr = stringify(re.sub(VAR_PATTERN, self.handle_var, text.replace('%', '%%')), strip)

    def handle_var(self, match):
        # local @var
        if match['var']:
            self.refs.append(match['var'])
            return '%d'

        # captured $var
        type = match['type']
        name = match['name']
        expr = match['expr'] or name
        name = name or slugify(expr)

        fmt = VAR_TYPES[type or 'ident']
        self.refs.append(fmt.fmt_expr(self.add_var(
            ctype = fmt.ctype,
            expr  = fmt.wrap_expr(name, expr),
            name  = name,
        )))

        if fmt.ctype == 'ident_t':
            return commentify(name, self.strip) + fmt.fmtstr
        else:
            return fmt.fmtstr

# Represents an entire GLSL block
class GLSLBlock(Statement):
    def __init__(self, line):
        super().__init__(linenr=line.linenr)
        self.lines = []
        self.refs  = []
        self.append(line)

    def append(self, line):
        assert isinstance(line, GLSLLine)
        self.lines.append(line)
        self.refs += line.refs
        self.vars.merge(line.vars)

    def render(self):
        return GLSL_BLOCK_TEMPLATE.render(block=self)

# Represents a statement which can either take a single line or a block
class BlockStart(Statement):
    def __init__(self, multiline=False, **kwargs):
        super().__init__(**kwargs)
        self.multiline = multiline

    def add_brace(self, text):
        if self.multiline:
            text += ' {'
        return text

# Represents an @if
class IfCond(BlockStart):
    def __init__(self, cond, inner=False, **kwargs):
        super().__init__(**kwargs)
        self.cond = cond if inner else self.add_var('bool', expr=cond)

    def render(self):
        return self.add_brace(f'if ({self.cond})')

# Represents an @else
class Else(BlockStart):
    def __init__(self, closing, **kwargs):
        super().__init__(**kwargs)
        self.closing = closing

    def render(self):
        text = '} else' if self.closing else 'else'
        return self.add_brace(text)

# Represents a normal (integer) @for loop, or an (unsigned 8-bit) bitmask loop
class ForLoop(BlockStart):
    def __init__(self, var, op, bound, **kwargs):
        super().__init__(**kwargs)
        self.comps = op == ':'
        self.bound = self.add_var('uint8_t' if self.comps else 'int', expr=bound)
        self.var   = var
        self.op    = op

    def render(self):
        if self.comps:
            loopstart = f'uint8_t _mask = {self.bound}, {self.var}'
            loopcond  = f'_mask && ({self.var} = __builtin_ctz(_mask), 1)'
            loopstep  = f'_mask &= ~(1u << {self.var})'
        else:
            loopstart = f'int {self.var} = 0'
            loopcond  = f'{self.var} {self.op} {self.bound}'
            loopstep  = f'{self.var}++'

        return self.add_brace(f'for ({loopstart}; {loopcond}; {loopstep})')

# Represents a @switch block
class Switch(Statement):
    def __init__(self, expr, **kwargs):
        super().__init__(**kwargs)
        self.expr = self.add_var('unsigned', expr=expr)

    def render(self):
        return f'switch ({self.expr}) {{'

# Represents a @case label
class Case(Statement):
    def __init__(self, label, **kwargs):
        super().__init__(**kwargs)
        self.label = label

    def render(self):
        return f'case {self.label}:'

# Represents a @default line
class Default(Statement):
    def render(self):
        return 'default:'

# Represents a @break line
class Break(Statement):
    def render(self):
        return 'break;'

# Represents a single closing brace
class EndBrace(Statement):
    def render(self):
        return '}'

# Shitty regex-based statement parser
PATTERN_IF  = re.compile(flags=re.VERBOSE, pattern=r'''
@\s*if\s*                       # '@if'
(?P<inner>@)?                   # optional leading @
\((?P<cond>.+)\)\s*             # (condition)
(?P<multiline>{)?\s*            # optional trailing {
$''')

PATTERN_ELSE = re.compile(flags=re.VERBOSE, pattern=r'''
@\s*(?P<closing>})?\s*          # optional leading }
else\s*                         # 'else'
(?P<multiline>{)?\s*            # optional trailing {
$''')

PATTERN_FOR = re.compile(flags=re.VERBOSE, pattern=r'''
@\s*for\s+\(                    # '@for' (
(?P<var>\w+)\s*                 # loop variable name
(?P<op>(?:\<=?|:))(?=[\w\s])\s* # '<', '<=' or ':', followed by \s or \w
(?P<bound>[^\s].*)\s*           # loop boundary expression
\)\s*(?P<multiline>{)?\s*       # ) and optional trailing {
$''')

PATTERN_SWITCH = re.compile(flags=re.VERBOSE, pattern=r'''
@\s*switch\s*                   # '@switch'
\((?P<expr>.+)\)\s*{            # switch expression
$''')

PATTERN_CASE = re.compile(flags=re.VERBOSE, pattern=r'''
@\s*case\s*                     # '@case'
(?P<label>[^:]+):?              # case label, optionally followed by :
$''')

PATTERN_BREAK   = r'@\s*break;?\s*$'
PATTERN_DEFAULT = r'@\s*default:?\s*$'
PATTERN_BRACE   = r'@\s*}\s*$'

PARSERS = {
    PATTERN_IF:         lambda r, **kw: IfCond(r['cond'], inner=r['inner'], multiline=r['multiline'], **kw),
    PATTERN_ELSE:       lambda r, **kw: Else(closing=r['closing'], multiline=r['multiline'], **kw),
    PATTERN_FOR:        lambda r, **kw: ForLoop(r['var'], r['op'], r['bound'], multiline=r['multiline'], **kw),
    PATTERN_SWITCH:     lambda r, **kw: Switch(r['expr'], **kw),
    PATTERN_CASE:       lambda r, **kw: Case(r['label'], **kw),
    PATTERN_BREAK:      lambda _, **kw: Break(**kw),
    PATTERN_DEFAULT:    lambda _, **kw: Default(**kw),
    PATTERN_BRACE:      lambda _, **kw: EndBrace(**kw),
}

def parse_line(text_orig, strip, **kwargs):
    # skip empty lines
    text = text_orig.strip()
    if not text:
        return None
    if text.lstrip().startswith('@'):
        # try parsing as statement
        for pat, fun in PARSERS.items():
            if res := re.match(pat, text):
                return fun(res, **kwargs)
        # return generic error for unrecognized statements
        raise SyntaxError('Syntax error in directive: ' + text.lstrip())
    else:
        # default to literal GLSL line
        return GLSLLine(text_orig, strip, **kwargs)

Statement.parse = parse_line



#!/usr/bin/env python3

import re

from variables import Var
from templates import *
from statement import *

PATTERN_PRAGMA  = re.compile(flags=re.VERBOSE, pattern=r'''
\s*\#\s*pragma\s+               # '#pragma'
(?P<pragma>(?:                  # pragma name
    GLSL[PHF]?
))\s*
(?P<rest>.*)$                   # rest of line (pragma body)
''')

# Represents a single #pragma macro
class Macro(object):
    PRAGMAS = {
        'GLSL':  'SH_BUF_BODY',
        'GLSLP': 'SH_BUF_PRELUDE',
        'GLSLH': 'SH_BUF_HEADER',
        'GLSLF': 'SH_BUF_FOOTER',
    }

    def __init__(self, linenr=0, type='GLSL'):
        self.linenr = linenr
        self.buf    = Macro.PRAGMAS[type]
        self.name   = '_glsl_' + str(linenr)
        self.body   = []    # list of statements
        self.last   = None  # previous GLSLBlock (if unterminated)
        self.vars   = VarSet()

    def needs_single_line(self):
        if not self.body:
            return False
        prev = self.body[-1]
        return isinstance(prev, BlockStart) and not prev.multiline

    def push_line(self, line):
        self.vars.merge(line.vars)

        if isinstance(line, GLSLLine):
            if self.last:
                self.last.append(line)
            elif self.needs_single_line():
                self.body.append(GLSLBlock(line))
            else:
                # start new GLSL block
                self.last = GLSLBlock(line)
                self.body.append(self.last)
        else:
            self.body.append(line)
            self.last = None

    def render_struct(self):
        return STRUCT_TEMPLATE.render(macro=self)

    def render_call(self):
        return CALL_TEMPLATE.render(macro=self)

    def render_fun(self):
        return FUNCTION_TEMPLATE.render(macro=self, Var=Var)

    # yields output lines
    @staticmethod
    def process_file(lines, strip=False):
        macro = None
        macros = []

        for linenr, line_orig in enumerate(lines, start=1):
            line = line_orig.rstrip()

            # Strip leading spaces, due to C indent. Skip first pragma line.
            if macro and leading_spaces is None:
                leading_spaces = len(line) - len(line.lstrip())

            # check for start of macro
            if not macro:
                leading_spaces = None
                if result := re.match(PATTERN_PRAGMA, line):
                    macro = Macro(linenr, type=result['pragma'])
                    line = result['rest'] # strip pragma prefix

            if macro:
                if leading_spaces:
                    line = re.sub(f'^\s{{1,{leading_spaces}}}', '', line)
                if more_lines := line.endswith('\\'):
                    line = line[:-1]
                if statement := Statement.parse(line, strip=strip, linenr=linenr):
                    macro.push_line(statement)
                if more_lines:
                    continue # stay in macro
                else:
                    yield macro.render_call()
                    yield '#line {}\n'.format(linenr + 1)
                    macros.append(macro)
                    macro = None
            else:
                yield line_orig

        if macros:
            yield '\n// Auto-generated template functions:'
        for macro in macros:
            yield macro.render_fun()



#!/usr/bin/env python3
#
# This file is part of libplacebo.
#
# libplacebo is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# libplacebo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with libplacebo.  If not, see <http://www.gnu.org/licenses/>.

import os.path
import re
import sys
import xml.etree.ElementTree as ET

try:
    import jinja2
except ModuleNotFoundError:
    print('Module \'jinja2\' not found, please install \'python3-Jinja2\' or '
          'an equivalent package on your system! Alternatively, run '
          '`git submodule update --init` followed by `meson --wipe`.',
          file=sys.stderr)
    sys.exit(1)

TEMPLATE = jinja2.Environment(
    loader = jinja2.FileSystemLoader(searchpath=os.path.dirname(__file__)),
    trim_blocks=True,
).get_template('utils_gen.c.j2')

class Obj(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

class VkXML(ET.ElementTree):
    def blacklist_block(self, req):
        for t in req.iterfind('type'):
            self.blacklist_types.add(t.attrib['name'])
        for e in req.iterfind('enum'):
            self.blacklist_enums.add(e.attrib['name'])

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.blacklist_types = set()
        self.blacklist_enums = set()

        for f in self.iterfind('feature'):
            # Feature block for non-Vulkan API
            if not 'vulkan' in f.attrib['api'].split(','):
                for r in f.iterfind('require'):
                    self.blacklist_block(r)

        for e in self.iterfind('extensions/extension'):
            # Entire extension is unsupported on vulkan or platform-specifid
            if not 'vulkan' in e.attrib['supported'].split(',') or 'platform' in e.attrib:
                for r in e.iterfind('require'):
                    self.blacklist_block(r)
                continue

            # Only individual <require> blocks are API-specific
            for r in e.iterfind('require[@api]'):
                if not 'vulkan' in r.attrib['api'].split(','):
                    self.blacklist_block(r)

    def findall_enum(self, name):
        for e in self.iterfind('enums[@name="{0}"]/enum'.format(name)):
            if not 'alias' in e.attrib:
                if not e.attrib['name'] in self.blacklist_enums:
                    yield e
        for e in self.iterfind('.//enum[@extends="{0}"]'.format(name)):
            if not 'alias' in e.attrib:
                if not e.attrib['name'] in self.blacklist_enums:
                    yield e

    def findall_type(self, category):
        for t in self.iterfind('types/type[@category="{0}"]'.format(category)):
            name = t.attrib.get('name') or t.find('name').text
            if name in self.blacklist_types:
                continue
            yield t


def get_vkenum(registry, enum):
    for e in registry.findall_enum(enum):
        yield e.attrib['name']

def get_vkobjects(registry):
    for t in registry.findall_type('handle'):
        if 'objtypeenum' in t.attrib:
            yield Obj(enum = t.attrib['objtypeenum'],
                      name = t.find('name').text)

def get_vkstructs(registry):
    for t in registry.findall_type('struct'):
        stype = None
        for m in t.iterfind('member'):
            if m.find('name').text == 'sType':
                stype = m
                break

        if stype is not None and 'values' in stype.attrib:
            yield Obj(stype = stype.attrib['values'],
                      name = t.attrib['name'])

def get_vkaccess(registry):
    access = Obj(read = 0, write = 0)
    for e in registry.findall_enum('VkAccessFlagBits2'):
        if '_READ_' in e.attrib['name']:
            access.read |= 1 << int(e.attrib['bitpos'])
        if '_WRITE_' in e.attrib['name']:
            access.write |= 1 << int(e.attrib['bitpos'])
    return access

def get_vkexts(registry):
    for e in registry.iterfind('extensions/extension'):
        promoted_ver = None
        if res := re.match(r'VK_VERSION_(\d)_(\d)', e.attrib.get('promotedto', '')):
            promoted_ver = 'VK_API_VERSION_{0}_{1}'.format(res[1], res[2])
        yield Obj(name = e.attrib['name'],
                  promoted_ver = promoted_ver)

def get_vkfeatures(registry):
    structs = [];
    featuremap = {}; # features -> [struct]
    for t in registry.findall_type('struct'):
        sname = t.attrib['name']
        is_base = sname == 'VkPhysicalDeviceFeatures'
        extends = t.attrib.get('structextends', [])
        if is_base:
            sname = 'VkPhysicalDeviceFeatures2'
            stype = 'VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2'
        elif not 'VkPhysicalDeviceFeatures2' in extends:
            continue

        features = []
        for f in t.iterfind('member'):
            if f.find('type').text == 'VkStructureType':
                stype = f.attrib['values']
            elif f.find('type').text == 'VkBool32':
                fname = f.find('name').text
                if is_base:
                    fname = 'features.' + fname
                features.append(Obj(name = fname))

        core_ver = None
        if res := re.match(r'VkPhysicalDeviceVulkan(\d)(\d)Features', sname):
            core_ver = 'VK_API_VERSION_{0}_{1}'.format(res[1], res[2])

        struct = Obj(name       = sname,
                     stype      = stype,
                     core_ver   = core_ver,
                     is_base    = is_base,
                     features   = features)

        structs.append(struct)
        for f in features:
            featuremap.setdefault(f.name, []).append(struct)

    for s in structs:
        for f in s.features:
            f.replacements = featuremap[f.name]
            core_ver = next(( r.core_ver for r in f.replacements if r.core_ver ), None)
            for r in f.replacements:
                if not r.core_ver:
                    r.max_ver = core_ver

    yield from structs

def find_registry_xml(datadir):
    registry_paths = [
        '{0}/vulkan/registry/vk.xml'.format(datadir),
        '$MINGW_PREFIX/share/vulkan/registry/vk.xml',
        '%VULKAN_SDK%/share/vulkan/registry/vk.xml',
        '$VULKAN_SDK/share/vulkan/registry/vk.xml',
        '/usr/share/vulkan/registry/vk.xml',
    ]

    for p in registry_paths:
        path = os.path.expandvars(p)
        if os.path.isfile(path):
            print('Found vk.xml: {0}'.format(path))
            return path

    print('Could not find the vulkan registry (vk.xml), please specify its '
          'location manually using the -Dvulkan-registry=/path/to/vk.xml '
          'option!', file=sys.stderr)
    sys.exit(1)

if __name__ == '__main__':
    assert len(sys.argv) == 4
    datadir = sys.argv[1]
    xmlfile = sys.argv[2]
    outfile = sys.argv[3]

    if not xmlfile or xmlfile == '':
        xmlfile = find_registry_xml(datadir)

    registry = VkXML(ET.parse(xmlfile))
    with open(outfile, 'w') as f:
        f.write(TEMPLATE.render(
            vkresults = get_vkenum(registry, 'VkResult'),
            vkformats = get_vkenum(registry, 'VkFormat'),
            vkspaces  = get_vkenum(registry, 'VkColorSpaceKHR'),
            vkhandles = get_vkenum(registry, 'VkExternalMemoryHandleTypeFlagBits'),
            vkalphas  = get_vkenum(registry, 'VkCompositeAlphaFlagBitsKHR'),
            vktransforms = get_vkenum(registry, 'VkSurfaceTransformFlagBitsKHR'),
            vkobjects = get_vkobjects(registry),
            vkstructs = get_vkstructs(registry),
            vkaccess = get_vkaccess(registry),
            vkexts = get_vkexts(registry),
            vkfeatures = get_vkfeatures(registry),
        ))
